=== modified file 'NEWS'
--- NEWS	2006-10-12 08:59:43 +0000
+++ NEWS	2006-10-13 06:30:29 +0000
@@ -3,6 +3,9 @@
   IMPROVEMENTS:
     * ``bzr help commands`` output is now shorter (Aaron Bentley)
 
+    * WSGI-compatible HTTP smart server.  See ``doc/http_smart_server.txt``.
+      (Andrew Bennetts)
+
   INTERNALS:
 
     * ``TestCase.run_bzr``, ``run_bzr_captured``, and ``run_bzr_subprocess``

=== modified file 'bzrlib/tests/test_wsgi.py'
--- bzrlib/tests/test_wsgi.py	2006-10-11 06:54:21 +0000
+++ bzrlib/tests/test_wsgi.py	2006-10-13 06:30:29 +0000
@@ -20,7 +20,7 @@
 
 from bzrlib import tests
 from bzrlib.transport.http import wsgi
-from bzrlib.transport import memory
+from bzrlib.transport import chroot, memory
 
 
 class TestWSGI(tests.TestCase):
@@ -69,11 +69,13 @@
         self.headers = headers
 
     def test_construct(self):
-        wsgi.SmartWSGIApp(None)
+        app = wsgi.SmartWSGIApp(FakeTransport())
+        self.assertIsInstance(
+            app.backing_transport, chroot.ChrootTransportDecorator)
 
     def test_http_get_rejected(self):
         # GET requests are rejected.
-        app = wsgi.SmartWSGIApp(None)
+        app = wsgi.SmartWSGIApp(FakeTransport())
         environ = self.build_environ({'REQUEST_METHOD': 'GET'})
         iterable = app(environ, self.start_response)
         self.read_response(iterable)
@@ -198,6 +200,29 @@
         self.assertEqual('200 OK', self.status)
         self.assertEqual('error\x01incomplete request\n', response)
 
+    def test_chrooting(self):
+        # Show that requests that try to access things outside of the base
+        # really will get intercepted by the ChrootTransportDecorator.
+        transport = memory.MemoryTransport()
+        transport.mkdir('foo')
+        transport.put_bytes('foo/bar', 'this is foo/bar')
+        wsgi_app = wsgi.SmartWSGIApp(transport.clone('foo'))
+
+        smart_request = StringIO('mkdir\x01/bad file\x01\n0\ndone\n')
+        environ = self.build_environ({
+            'REQUEST_METHOD': 'POST',
+            'CONTENT_LENGTH': len(smart_request.getvalue()),
+            'wsgi.input': smart_request,
+            'bzrlib.relpath': '.',
+        })
+        iterable = wsgi_app(environ, self.start_response)
+        response = self.read_response(iterable)
+        self.assertEqual('200 OK', self.status)
+        self.assertEqual(
+            "error\x01Path '/bad file' is not a child of "
+            "path 'memory:///foo/'\n",
+            response)
+
 
 class FakeRequest(object):
     
@@ -218,6 +243,10 @@
 
     def __init__(self):
         self.calls = []
+        self.base = 'fake:///'
+
+    def abspath(self, relpath):
+        return 'fake:///' + relpath
 
     def clone(self, relpath):
         self.calls.append(('clone', relpath))

=== modified file 'bzrlib/transport/http/wsgi.py'
--- bzrlib/transport/http/wsgi.py	2006-10-11 06:54:21 +0000
+++ bzrlib/transport/http/wsgi.py	2006-10-13 06:30:29 +0000
@@ -22,7 +22,7 @@
 
 from cStringIO import StringIO
 
-from bzrlib.transport import get_transport, smart
+from bzrlib.transport import chroot, get_transport, smart
 from bzrlib.urlutils import local_path_to_url
     
 
@@ -86,7 +86,12 @@
         :param backing_transport: a transport.  Requests will be processed
             relative to this transport.
         """
-        self.backing_transport = backing_transport
+        # Use a ChrootTransportDecorator so that this web application won't
+        # accidentally let people access locations they shouldn't.
+        # e.g. consider a smart server request for "get /etc/passwd" or
+        # something.
+        self.backing_transport = chroot.ChrootTransportDecorator(
+            'chroot+' + backing_transport.base, _decorated=backing_transport)
 
     def __call__(self, environ, start_response):
         """WSGI application callable."""
@@ -96,7 +101,6 @@
 
         relpath = environ['bzrlib.relpath']
         transport = self.backing_transport.clone(relpath)
-        #assert transport.base.startswith(self.backing_transport.base)
         out_buffer = StringIO()
         smart_protocol_request = self.make_request(transport, out_buffer.write)
         request_data_length = int(environ['CONTENT_LENGTH'])

